from .core import TransformConverterBase, import_type, parameter_to_value


class ProjectionConverter(TransformConverterBase):
    """
    ASDF support for serializing most projection models.
    An instance of this class must be created for each
    projection.

    Parameters
    ----------
    tags : list of str
        Tag patterns.
    pix2sky_type_name : str
        Fully-qualified type name of the projection's Pix2Sky model.
    sky2pix_type_name : str
        Fully-qualified type name of the projection's Sky2Pix model.
    """
    def __init__(self, tags, pix2sky_type_name, sky2pix_type_name):
        self._tags = tags
        self._pix2sky_type_name = pix2sky_type_name
        self._sky2pix_type_name = sky2pix_type_name
        self._pix2sky_type = None
        self._sky2pix_type = None

    @property
    def tags(self):
        return self._tags

    @property
    def types(self):
        return [self._pix2sky_type_name, self._sky2pix_type_name]

    @property
    def pix2sky_type(self):
        # Delay import until the model class is needed to improve speed
        # of loading the extension.
        if self._pix2sky_type is None:
            self._pix2sky_type = import_type(self._pix2sky_type_name)
        return self._pix2sky_type

    @property
    def sky2pix_type(self):
        # Delay import until the model class is needed to improve speed
        # of loading the extension.
        if self._sky2pix_type is None:
            self._sky2pix_type = import_type(self._sky2pix_type_name)
        return self._sky2pix_type

    def to_yaml_tree_transform(self, model, tag, ctx):
        if isinstance(model, self.pix2sky_type):
            direction = "pix2sky"
        elif isinstance(model, self.sky2pix_type):
            direction = "sky2pix"
        else:
            raise TypeError("Unrecognized projection model type")

        node = {p: parameter_to_value(getattr(model, p)) for p in model.param_names}
        node["direction"] = direction

        return node

    def from_yaml_tree_transform(self, node, tag, ctx):
        if node["direction"] == "pix2sky":
            model_type = self.pix2sky_type
        elif node["direction"] == "sky2pix":
            model_type = self.sky2pix_type
        else:
            raise ValueError("Unrecognized projection direction: " + node["direction"])

        model_kwargs = {}
        for param in model_type.param_names:
            if param in node:
                model_kwargs[param] = node[param]
        return model_type(**model_kwargs)



import pytest

import asdf
import astropy.units as u
from astropy.coordinates import SkyCoord, FK4, Galactic, ICRS, Longitude
import numpy as np

from asdf_astropy.testing.helpers import assert_sky_coord_equal


def create_sky_coords():
    # These are cribbed directly from the Examples section of
    # https://docs.astropy.org/en/stable/api/astropy.coordinates.SkyCoord.html
    return [
        # Defaults to ICRS frame
        SkyCoord(10, 20, unit="deg"),
        # Vector of 3 coords
        SkyCoord([1, 2, 3], [-30, 45, 8], frame="icrs", unit="deg"),
        # FK4 frame
        SkyCoord(
            ["1:12:43.2 +1:12:43", "1 12 43.2 +1 12 43"],
            frame=FK4,
            unit=(u.deg, u.hourangle),
            obstime="J1992.21",
        ),
        # Galactic frame
        SkyCoord("1h12m43.2s +1d12m43s", frame=Galactic),
        SkyCoord(frame="galactic", l="1h12m43.2s", b="+1d12m43s"),
        # With ra and dec
        SkyCoord(
            Longitude([1, 2, 3], unit=u.deg),
            np.array([4.5, 5.2, 6.3]) * u.deg,
            frame="icrs",
        ),
        SkyCoord(
            frame=ICRS,
            ra=Longitude([1, 2, 3], unit=u.deg),
            dec=np.array([4.5, 5.2, 6.3]) * u.deg,
            obstime="2001-01-02T12:34:56"
        ),
        # With overridden frame defaults
        SkyCoord(FK4(1 * u.deg, 2 * u.deg), obstime="J2010.11", equinox="B1965"),
        # Cartesian
        SkyCoord(w=0, u=1, v=2, unit="kpc", frame="galactic", representation_type="cartesian"),
        # Vector frame
        SkyCoord([ICRS(ra=1 * u.deg, dec=2 * u.deg), ICRS(ra=3 * u.deg, dec=4 * u.deg)]),
        # 2D obstime
        SkyCoord([1, 2], [3, 4], [5, 6], unit="deg,deg,m", frame="fk4", obstime=["J1990.5", "J1991.5"]),
        # Radial velocity
        SkyCoord(ra=1 * u.deg, dec=2 * u.deg, radial_velocity=10 * u.km / u.s),
        # Proper motion
        SkyCoord(ra=1 * u.deg, dec=2 * u.deg, pm_ra_cosdec=2 * u.mas / u.yr, pm_dec=1 * u.mas / u.yr),
    ]


@pytest.mark.parametrize("coord", create_sky_coords())
def test_serialization(coord, tmp_path):
    file_path = tmp_path / "test.asdf"

    with asdf.AsdfFile() as af:
        af["coord"] = coord
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert_sky_coord_equal(af["coord"], coord)


def test_serializaton_extra_attribute(tmp_path):
    coord = SkyCoord(10 * u.deg, 20 * u.deg, equinox="2011-01-01T00:00", frame="fk4").transform_to("icrs")

    file_path = tmp_path / "test.asdf"

    with asdf.AsdfFile() as af:
        af["coord"] = coord
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert_sky_coord_equal(af["coord"], coord)
        assert hasattr(af["coord"], "equinox")



from pathlib import Path
import sys

import yaml
import asdf


def test_resources():
    resources_root = Path(__file__).parent.parent / "resources"
    resource_manager = asdf.get_config().resource_manager

    for resource_path in resources_root.glob("**/*.yaml"):
        with resource_path.open("rb") as f:
            resource_content = f.read()
        resource = yaml.safe_load(resource_content)
        resource_uri = resource["id"]
        assert resource_manager[resource_uri] == resource_content


def test_manifests():
    manifests_root = (
        Path(__file__).parent.parent / "resources" / "manifests"
    )
    resource_manager = asdf.get_config().resource_manager

    for manifest_path in manifests_root.glob("*.yaml"):
        with manifest_path.open("rb") as f:
            manifest_content = f.read()
        manifest = yaml.safe_load(manifest_content)

        manifest_schema = asdf.schema.load_schema(
            "asdf://asdf-format.org/core/schemas/extension_manifest-1.0.0"
        )

        # The manifest must be valid against its own schema:
        asdf.schema.validate(manifest, schema=manifest_schema)

        for tag_definition in manifest["tags"]:
            # The tag's schema must be available:
            assert tag_definition["schema_uri"] in resource_manager


def test_extensions():
    package_and_uri_pairs = {(e.package_name, e.extension_uri) for e in asdf.get_config().extensions}

    assert ("asdf-astropy", "asdf://astropy.org/astropy/extensions/astropy-1.0.0") in package_and_uri_pairs
    assert ("asdf-astropy", "asdf://asdf-format.org/transform/extensions/transform-1.0.0") in package_and_uri_pairs
    assert ("asdf-astropy", "asdf://asdf-format.org/transform/extensions/transform-1.1.0") in package_and_uri_pairs
    assert ("asdf-astropy", "asdf://asdf-format.org/transform/extensions/transform-1.2.0") in package_and_uri_pairs
    assert ("asdf-astropy", "asdf://asdf-format.org/transform/extensions/transform-1.3.0") in package_and_uri_pairs
    assert ("asdf-astropy", "asdf://asdf-format.org/transform/extensions/transform-1.4.0") in package_and_uri_pairs
    assert ("asdf-astropy", "asdf://asdf-format.org/transform/extensions/transform-1.5.0") in package_and_uri_pairs


_ASTROPY_MODULES = [
    "astropy.coordinates",
    "astropy.io",
    "astropy.modeling",
    "astropy.table",
    "astropy.time",
    "astropy.units",
]


def test_no_astropy_import():
    """
    Confirm that none of the ASDF plugins import astropy modules
    at import time.
    """

    keys = [k for k in sys.modules.keys() if k.startswith("asdf_astropy") or any(k.startswith(m) for m in _ASTROPY_MODULES)]
    for key in keys:
        del sys.modules[key]

    from asdf_astropy import integration
    integration.get_resource_mappings()
    integration.get_extensions()

    assert not any(k for k in sys.modules.keys() if any(k.startswith(m) for m in _ASTROPY_MODULES))



import abc
import warnings

import numpy as np
from asdf.extension import Converter

from ..utils import import_type


def parameter_to_value(param):
    """
    Convert a model parameter to a Quantity or number,
    depending on the presence of a unit.

    Parameters
    ----------
    param : astropy.modeling.Parameter

    Returns
    -------
    astropy.units.Quantity or float
    """
    from astropy import units as u

    if param.unit is not None:
        return u.Quantity(param)
    else:
        return param.value


# One converter, UnitsMappingConverter, does not inherit
# this class.  When adding features here consider also
# updating UnitsMappingConverter.
# This class is used by other packages, e.g., gwcs, to implement
# converters for custom models.  Keep that in mind when modifying
# this code.
class TransformConverterBase(Converter):
    """
    ABC for transform/model converters.  Handles common
    properties after concrete converter sets model-specific
    properties.
    """
    @abc.abstractmethod
    def to_yaml_tree_transform(self, model, tag, ctx):
        """
        Convert a model's parameters into a dict suitable
        for ASDF serialization.  Common model properties
        such as name and inverse will be handled by this
        base class.

        Parameters
        ----------
        model : astropy.modeling.Model
            The model instance to convert.
        tag : str
            The tag identifying the YAML type that `model` should be
            converted into.
        ctx : asdf.asdf.SerializationContext
            The context of the current serialization request.

        Returns
        -------
        dict
            ASDF node.
        """
        pass  # pragma: no cover

    @abc.abstractmethod
    def from_yaml_tree_transform(self, node, tag, ctx):
        """
        Convert an ASDF node into an instance of the appropriate
        model class.  The implementing class need only instantiate
        the model and set parameter values; common model properties
        such as name and inverse will be handled by this base class.

        Parameters
        ----------
        node : dict
            The ASDF node to convert.
        tag : str
            The tag identifying the YAML type of the node.
        ctx : asdf.asdf.SerializationContext
            The context of the current serialization request.

        Returns
        -------
        astropy.modeling.Model
            The resulting model instance.
        """
        pass  # pragma: no cover

    def to_yaml_tree(self, model, tag, ctx):
        from astropy.modeling.core import CompoundModel

        node = self.to_yaml_tree_transform(model, tag, ctx)

        if model.name is not None:
            node["name"] = model.name

        node["inputs"] = list(model.inputs)
        node["outputs"] = list(model.outputs)

        # Don't bother serializing analytic inverses provided
        # by the model:
        if getattr(model, "_user_inverse", None) is not None:
            node["inverse"] = model._user_inverse

        self._serialize_bounding_box(model, node)

        # model / parameter constraints
        if not isinstance(model, CompoundModel):
            fixed_nondefaults = {k: f for k, f in model.fixed.items() if f}
            if fixed_nondefaults:
                node['fixed'] = fixed_nondefaults
            bounds_nondefaults = {k: b for k, b in model.bounds.items() if any(b)}
            if bounds_nondefaults:
                node['bounds'] = bounds_nondefaults

        return node

    def _serialize_bounding_box(self, model, node):
        from packaging.version import Version
        import astropy

        if Version(astropy.__version__) > Version('4.999.999'):
            self._serialize_bounding_box_astropy_5(model, node)
        else:
            self._serialize_bounding_box_astropy_4(model, node)

    def _serialize_bounding_box_astropy_4(self, model, node):
        try:
            bb = model.bounding_box
        except NotImplementedError:
            bb = None

        if bb is not None:
            if model.n_inputs == 1:
                bb = list(bb)
            else:
                bb = [list(item) for item in model.bounding_box]
            node['bounding_box'] = bb

    def _serialize_bounding_box_astropy_5(self, model, node):
        from astropy.modeling.bounding_box import ModelBoundingBox, CompoundBoundingBox

        try:
            bb = model.bounding_box
        except NotImplementedError:
            bb = None

        if isinstance(bb, ModelBoundingBox):
            bb = bb.bounding_box(order='C')

            if model.n_inputs == 1:
                bb = list(bb)
            else:
                bb = [list(item) for item in bb]
            node['bounding_box'] = bb

        elif isinstance(bb, CompoundBoundingBox):
            selector_args = [[sa.index, sa.ignore] for sa in bb.selector_args]
            node['selector_args'] = selector_args
            node['cbbox_keys'] = list(bb.bounding_boxes.keys())

            bounding_boxes = list(bb.bounding_boxes.values())
            if len(model.inputs) - len(selector_args) == 1:
                node['cbbox_values'] = [list(sbbox.bounding_box()) for sbbox in bounding_boxes]
            else:
                node['cbbox_values'] = [[list(item) for item in sbbox.bounding_box()
                                         if np.isfinite(item[0])] for sbbox in bounding_boxes]

    def from_yaml_tree(self, node, tag, ctx):
        model = self.from_yaml_tree_transform(node, tag, ctx)

        if 'name' in node:
            model.name = node['name']

        self._deserialize_bounding_box(model, node)

        if 'inputs' in node:
            model.inputs = tuple(node['inputs'])

        if 'outputs' in node:
            model.outputs = tuple(node['outputs'])

        param_and_model_constraints = {}
        for constraint in ['fixed', 'bounds']:
            if constraint in node:
                param_and_model_constraints[constraint] = node[constraint]
        model._initialize_constraints(param_and_model_constraints)

        yield model

        if 'inverse' in node:
            model.inverse = node['inverse']

    def _deserialize_bounding_box(self, model, node):
        from packaging.version import Version
        import astropy

        if Version(astropy.__version__) > Version('4.999.999'):
            self._deserialize_bounding_box_astropy_5(model, node)
        else:
            self._deserialize_bounding_box_astropy_4(model, node)

    def _deserialize_bounding_box_astropy_4(self, model, node):
        if 'bounding_box' in node:
            model.bounding_box = node['bounding_box']
        elif 'selector_args' in node:
            warnings.warn("This version of astropy does not support compound bounding boxes.")

    def _deserialize_bounding_box_astropy_5(self, model, node):
        from astropy.modeling.bounding_box import CompoundBoundingBox

        if 'bounding_box' in node:
            model.bounding_box = node['bounding_box']
        elif 'selector_args' in node:
            cbbox_keys = [tuple(key) for key in node['cbbox_keys']]
            bbox_dict = dict(zip(cbbox_keys, node['cbbox_values']))

            selector_args = node['selector_args']
            model.bounding_box = CompoundBoundingBox.validate(model, bbox_dict, selector_args)


class SimpleTransformConverter(TransformConverterBase):
    """
    Class for converters that serialize all of a model's parameters
    and do not require special behavior based on tag version.

    Parameters
    ----------
    tags : list of str
        Tag patterns.

    model_type_name
        Fully-qualified model type name.
    """
    def __init__(self, tags, model_type_name):
        self._tags = tags
        self._model_type_name = model_type_name
        self._model_type = None

    @property
    def tags(self):
        return self._tags

    @property
    def types(self):
        return [self._model_type_name]

    @property
    def model_type(self):
        # Delay import until the model class is needed to improve speed
        # of loading the extension.
        if self._model_type is None:
            self._model_type = import_type(self._model_type_name)
        return self._model_type

    def to_yaml_tree_transform(self, model, tag, ctx):
        return {p: parameter_to_value(getattr(model, p)) for p in model.param_names}

    def from_yaml_tree_transform(self, node, tag, ctx):
        model_type = self.model_type
        model_kwargs = {}
        for param in model_type.param_names:
            if param in node:
                model_kwargs[param] = node[param]
        return model_type(**model_kwargs)



import importlib


def import_type(type_name):
    """
    Import a Python type from its fully-qualified name.
    For example, when this method is called with 'builtins.str'
    it will return the `str` type.

    Parameters
    ----------
    type_name : str
        Type name, with module.

    Returns
    -------
    type
    """
    module_name, class_name = type_name.rsplit(".", 1)
    return getattr(importlib.import_module(module_name), class_name)



import pytest

import asdf
from astropy import units as u
from astropy.units import Quantity
from astropy.coordinates import EarthLocation, Longitude, Latitude
from astropy.coordinates.earth import ELLIPSOIDS

from asdf_astropy.testing.helpers import assert_earth_location_equal


def create_earth_locations():
    longitude = Longitude([0., 45., 90., 135., 180., -180, -90, -45], u.deg, wrap_angle=180 * u.deg)
    latitude = Latitude([+0., 30., 60., +90., -90., -60., -30., 0.], u.deg)
    height = Quantity([0.1, 0.5, 1.0, -0.5, -1.0, +4.2, -11., -.1], u.m)
    position = (longitude, latitude, height)

    result = [
        EarthLocation(lat=34.4900 * u.deg, lon=-104.221800 * u.deg, height=40 * u.km),
        EarthLocation(*EarthLocation.from_geodetic(*position).to_geocentric()),
    ]

    result.extend([EarthLocation.from_geodetic(*position, ellipsoid=e) for e in ELLIPSOIDS])

    return result


@pytest.mark.parametrize("earth_location", create_earth_locations())
def test_serialization(earth_location, tmp_path):
    file_path = tmp_path / "test.asdf"
    with asdf.AsdfFile() as af:
        af["earth_location"] = earth_location
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert (af["earth_location"] == earth_location).all()


@pytest.fixture
def builtin_site_registry():
    orig_sites = getattr(EarthLocation, "_site_registry", None)
    EarthLocation._get_site_registry(force_builtin=True)
    yield
    EarthLocation._site_registry = orig_sites


def test_earthlocation_site(tmp_path, builtin_site_registry):
    earth_location = EarthLocation.of_site("greenwich")

    file_path = tmp_path / "test.asdf"
    with asdf.AsdfFile() as af:
        af["earth_location"] = earth_location
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert_earth_location_equal(af["earth_location"], earth_location)





import pytest

import asdf
from astropy import units as u
from astropy.coordinates import (
    Angle,
    CartesianRepresentation,
    CIRS,
    FK4,
    FK4NoETerms,
    FK5,
    Galactic,
    Galactocentric,
    GCRS,
    ICRS,
    ITRS,
    Latitude,
    Longitude,
    PrecessedGeocentric,
    SphericalRepresentation,
)
from astropy.time import Time
import numpy as np

from asdf_astropy.testing.helpers import assert_frame_equal


def create_frames():
    test_data = {
       "ra": 1 * u.deg,
        "dec": 2 * u.deg,
    }

    return [
        CIRS(**test_data),
        CIRS(**test_data, obstime=Time("J2005")),
        FK4(**test_data),
        FK4(**test_data, obstime=Time("B1950")),
        FK4NoETerms(**test_data),
        FK4NoETerms(**test_data, obstime= Time("J1975")),
        FK5(**test_data),
        FK5(**test_data, equinox="J2005"),
        FK5(**test_data, equinox="2011-01-01T00:00:00"),
        Galactic(l=47.37 * u.degree, b=+6.32 * u.degree),
        Galactocentric(
            x=np.linspace(-10., 10., 100) * u.kpc,
            y=np.linspace(-10., 10., 100) * u.kpc,
            z=np.zeros(100) * u.kpc,
            z_sun=15 * u.pc
        ),
        GCRS(**test_data),
        GCRS(**test_data, obsgeoloc=CartesianRepresentation([1, 2, 3], unit=u.m)),
        ICRS(**test_data),
        ICRS(ra=Longitude(25, unit=u.deg), dec=Latitude(45, unit=u.deg)),
        ICRS(ra=Longitude(25, unit=u.deg, wrap_angle=Angle(1.5, unit=u.rad)), dec=Latitude(45, unit=u.deg)),
        ICRS(ra=[0, 1, 2] * u.deg, dec=[3, 4, 5] * u.deg),
        ITRS(SphericalRepresentation(lon=12.3 * u.deg, lat=45.6 * u.deg, distance=1 * u.km)),
        PrecessedGeocentric(**test_data),
        PrecessedGeocentric(**test_data, equinox="B1975"),
    ]


@pytest.mark.parametrize("frame", create_frames())
def test_serialization(frame, tmp_path):
    file_path = tmp_path / "test.asdf"
    with asdf.AsdfFile() as af:
        af["frame"] = frame
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert_frame_equal(af["frame"], frame)



from astropy.coordinates import EarthLocation
from astropy.tests.helper import assert_quantity_allclose
from astropy import units
from numpy.testing import assert_array_equal


def assert_earth_location_equal(a, b):
    __tracebackhide__ = True

    assert (a == b).all()


def assert_representation_equal(a, b):
    __tracebackhide__ = True

    assert type(a) is type(b)
    assert a.components == b.components
    for component in a.components:
        assert units.allclose(getattr(a, component), getattr(b, component))


def assert_sky_coord_equal(a, b):
    __tracebackhide__ = True

    assert a.is_equivalent_frame(b)
    assert a.representation_type is b.representation_type
    assert a.shape == b.shape

    assert_representation_equal(a.data, b.data)


def assert_frame_equal(a, b):
    __tracebackhide__ = True

    assert type(a) is type(b)

    if a is None:
        return

    assert_representation_equal(a.data, b.data)


def assert_spectral_coord_equal(a, b):
    __tracebackhide__ = True

    assert type(a) is type(b)
    assert_quantity_allclose(a.quantity, b.quantity)
    assert_frame_equal(a.observer, b.observer)
    assert_frame_equal(a.target, b.target)


def assert_time_equal(a, b):
    assert a.format == b.format
    assert a.scale == b.scale

    assert type(a.location) == type(b.location)
    if isinstance(a.location, EarthLocation):
        assert_earth_location_equal(a.location, b.location)
    else:
        assert a.location == b.location

    assert_array_equal(a, b)


def assert_time_delta_equal(a, b):
    assert_array_equal(a.jd, b.jd)
    assert_array_equal(a.jd2, b.jd2)
    assert_array_equal(a.sec, b.sec)


def assert_hdu_list_equal(a, b):
    assert len(a) == len(b)
    for hdu_a, hdu_b in zip(a, b):
        assert_array_equal(hdu_a.data, hdu_b.data)
        assert len(hdu_a.header.cards) == len(hdu_b.header.cards)
        for card_a, card_b in zip(hdu_a.header.cards, hdu_b.header.cards):
            assert tuple(card_a) == tuple(card_b)



from packaging.version import parse as parse_version
import numpy as np

from .core import TransformConverterBase

from ..helpers import parse_tag_version


class PolynomialConverter(TransformConverterBase):
    """
    ASDF support for serializing the 1D and 2D polynomial models.
    """
    # Schema versions prior to 1.2 included an unrelated "domain"
    # property.  We can't serialize the new domain values with those
    # versions because they don't validate.
    _DOMAIN_WINDOW_MIN_VERSION = parse_version("1.2.0")

    tags = ["tag:stsci.edu:asdf/transform/polynomial-*"]

    types = [
        "astropy.modeling.polynomial.Polynomial1D",
        "astropy.modeling.polynomial.Polynomial2D",
    ]

    def to_yaml_tree_transform(self, model, tag, ctx):
        from astropy.modeling.polynomial import Polynomial1D, Polynomial2D

        if isinstance(model, Polynomial1D):
            coefficients = np.array(model.parameters)
        elif isinstance(model, Polynomial2D):
            degree = model.degree
            coefficients = np.zeros((degree + 1, degree + 1))
            for i in range(degree + 1):
                for j in range(degree + 1):
                    if i + j < degree + 1:
                        name = "c" + str(i) + "_" + str(j)
                        coefficients[i, j] = getattr(model, name).value

        node = {"coefficients": coefficients}

        if parse_tag_version(tag) >= self._DOMAIN_WINDOW_MIN_VERSION:
            if model.n_inputs == 1:
                if model.domain is not None:
                    node["domain"] = model.domain
                if model.window is not None:
                    node["window"] = model.window
            else:
                if model.x_domain or model.y_domain is not None:
                    node["domain"] = (model.x_domain, model.y_domain)
                if model.x_window or model.y_window is not None:
                    node["window"] = (model.x_window, model.y_window)

        return node

    def from_yaml_tree_transform(self, node, tag, ctx):
        from astropy.modeling.polynomial import Polynomial1D, Polynomial2D

        coefficients = np.asarray(node["coefficients"])
        n_dim = coefficients.ndim

        if n_dim == 1:
            domain = node.get("domain", None)
            window = node.get("window", None)

            model = Polynomial1D(coefficients.size - 1,
                                 domain=domain, window=window)
            model.parameters = coefficients
        elif n_dim == 2:
            x_domain, y_domain = tuple(node.get("domain", (None, None)))
            x_window, y_window = tuple(node.get("window", (None, None)))
            shape = coefficients.shape
            degree = shape[0] - 1
            if shape[0] != shape[1]:
                raise TypeError("Coefficients must be an (n+1, n+1) matrix")

            coeffs = {}
            for i in range(shape[0]):
                for j in range(shape[0]):
                    if i + j < degree + 1:
                        name = "c" + str(i) + "_" + str(j)
                        coeffs[name] = coefficients[i, j]
            model = Polynomial2D(degree,
                                 x_domain=x_domain,
                                 y_domain=y_domain,
                                 x_window=x_window,
                                 y_window=y_window,
                                 **coeffs)
        else:
            raise NotImplementedError("astropy supports only 1D or 2D polynomial models")

        return model


class OrthoPolynomialConverter(TransformConverterBase):
    """
    ASDF support for serializing models that inherit
    OrthoPolyomialBase.
    """
    # Map of model class name to (polynomial type, number of dimensions) tuple:
    _CLASS_NAME_TO_POLY_INFO = {
        "Legendre1D": ("legendre", 1),
        "Legendre2D": ("legendre", 2),
        "Chebyshev1D": ("chebyshev", 1),
        "Chebyshev2D": ("chebyshev", 2),
        "Hermite1D": ("hermite", 1),
        "Hermite2D": ("hermite", 2),
    }

    _POLY_INFO_TO_CLASS_NAME = {v: k for k, v in _CLASS_NAME_TO_POLY_INFO.items()}

    tags = ["tag:stsci.edu:asdf/transform/ortho_polynomial-*"]

    types = [
        "astropy.modeling.polynomial.Legendre1D",
        "astropy.modeling.polynomial.Legendre2D",
        "astropy.modeling.polynomial.Chebyshev1D",
        "astropy.modeling.polynomial.Chebyshev2D",
        "astropy.modeling.polynomial.Hermite1D",
        "astropy.modeling.polynomial.Hermite2D",
    ]

    def to_yaml_tree_transform(self, model, tag, ctx):
        poly_type = self._CLASS_NAME_TO_POLY_INFO[model.__class__.__name__][0]
        if model.n_inputs == 1:
            coefficients = np.array(model.parameters)
        else:
            coefficients = np.zeros((model.x_degree + 1, model.y_degree + 1))
            for i in range(model.x_degree + 1):
                for j in range(model.y_degree + 1):
                    name = f"c{i}_{j}"
                    coefficients[i, j] = getattr(model, name).value

        node = {"polynomial_type": poly_type, "coefficients": coefficients}

        if model.n_inputs == 1:
            if model.domain is not None:
                node["domain"] = model.domain
            if model.window is not None:
                node["window"] = model.window
        else:
            if model.x_domain or model.y_domain is not None:
                node["domain"] = (model.x_domain, model.y_domain)
            if model.x_window or model.y_window is not None:
                node["window"] = (model.x_window, model.y_window)

        return node

    def from_yaml_tree_transform(self, node, tag, ctx):
        from astropy.modeling import polynomial

        coefficients = np.asarray(node["coefficients"])
        poly_type = node["polynomial_type"]
        n_dim = coefficients.ndim

        class_name = self._POLY_INFO_TO_CLASS_NAME[(poly_type, n_dim)]
        model_type = getattr(polynomial, class_name)

        coefficients = np.asarray(node["coefficients"])
        if n_dim == 1:
            domain = node.get("domain", None)
            window = node.get("window", None)
            model = model_type(coefficients.size - 1, domain=domain, window=window)
            model.parameters = coefficients
        elif n_dim == 2:
            x_domain, y_domain = tuple(node.get("domain", (None, None)))
            x_window, y_window = tuple(node.get("window", (None, None)))
            coeffs = {}
            shape = coefficients.shape
            x_degree = shape[0] - 1
            y_degree = shape[1] - 1
            for i in range(x_degree + 1):
                for j in range(y_degree + 1):
                    name = f"c{i}_{j}"
                    coeffs[name] = coefficients[i, j]
            model = model_type(
                x_degree,
                y_degree,
                x_domain=x_domain,
                y_domain=y_domain,
                x_window=x_window,
                y_window=y_window,
                **coeffs
            )
        else:
            raise NotImplementedError("astropy supports only 1D or 2D polynomial models")

        return model



import sys

from asdf.resource import DirectoryResourceMapping

if sys.version_info < (3, 9):
    import importlib_resources
else:
    import importlib.resources as importlib_resources


def get_resource_mappings():
    """
    Get the resource mapping instances for the astropy schemas
    and manifests.  This method is registered with the
    asdf.resource_mappings entry point.

    Returns
    -------
    list of collections.abc.Mapping
    """
    from . import resources
    resources_root = importlib_resources.files(resources)

    return [
        DirectoryResourceMapping(resources_root / "schemas", "http://astropy.org/schemas/astropy/", recursive=True),
        DirectoryResourceMapping(resources_root / "manifests", "asdf://astropy.org/astropy/manifests/"),
    ]


def get_extensions():
    """
    Get the extension instances for the various astropy
    extensions.  This method is registered with the
    asdf.extensions entry point.

    Returns
    -------
    list of asdf.extension.Extension
    """
    from . import extensions
    return [extensions.ASTROPY_EXTENSION, extensions.COORDINATES_EXTENSION] + extensions.TRANSFORM_EXTENSIONS + extensions.CORE_EXTENSIONS



# This file is used to configure the behavior of pytest when using the Astropy
# test infrastructure. It needs to live inside the package in order for it to
# get picked up when running the tests inside an interpreter using
# packagename.test

import os

import asdf
from astropy.version import version as astropy_version
import pytest

# For Astropy 3.0 and later, we can use the standalone pytest plugin
if astropy_version < '3.0':
    from astropy.tests.pytest_plugins import *  # noqa
    del pytest_report_header
    ASTROPY_HEADER = True
else:
    try:
        from pytest_astropy_header.display import PYTEST_HEADER_MODULES, TESTED_VERSIONS
        ASTROPY_HEADER = True
    except ImportError:
        ASTROPY_HEADER = False


def pytest_configure(config):

    if ASTROPY_HEADER:

        config.option.astropy_header = True

        # Customize the following lines to add/remove entries from the list of
        # packages for which version numbers are displayed when running the tests.
        PYTEST_HEADER_MODULES.pop('Pandas', None)
        PYTEST_HEADER_MODULES['scikit-image'] = 'skimage'

        from . import __version__
        packagename = os.path.basename(os.path.dirname(__file__))
        TESTED_VERSIONS[packagename] = __version__

# Uncomment the last two lines in this block to treat all DeprecationWarnings as
# exceptions. For Astropy v2.0 or later, there are 2 additional keywords,
# as follow (although default should work for most cases).
# To ignore some packages that produce deprecation warnings on import
# (in addition to 'compiler', 'scipy', 'pygments', 'ipykernel', and
# 'setuptools'), add:
#     modules_to_ignore_on_import=['module_1', 'module_2']
# To ignore some specific deprecation warning messages for Python version
# MAJOR.MINOR or later, add:
#     warnings_to_ignore_by_pyver={(MAJOR, MINOR): ['Message to ignore']}
# from astropy.tests.helper import enable_deprecations_as_exceptions  # noqa
# enable_deprecations_as_exceptions()


@pytest.fixture(autouse=True)
def remove_astropy_extensions():
    """
    Disable the old astropy extension so that it doesn't
    confuse our test results.
    """
    with asdf.config_context() as config:
        config.remove_extension(package="astropy")
        yield



import os

from astropy.tests.runner import TestRunner

__all__ = ['__version__', 'test']

try:
    from .version import version as __version__
except ImportError:
    __version__ = ''

# Create the test function for self test
test = TestRunner.make_test_runner_in(os.path.dirname(__file__))



import asdf
from asdf.tests.helpers import yaml_to_asdf
from astropy.coordinates import SkyCoord, EarthLocation
from astropy.table import Table, NdarrayMixin, QTable, Column, MaskedColumn
from astropy.time import Time, TimeDelta
import astropy.units as u
import numpy as np
from numpy.testing import assert_array_equal
import pytest

from asdf_astropy.testing.helpers import (
    assert_earth_location_equal,
    assert_sky_coord_equal,
    assert_time_equal,
    assert_time_delta_equal,
)


def assert_description_equal(a, b):
    if a in ("", None) and b in ("", None):
        return

    assert a == b


def assert_table_equal(a, b):
    assert type(a) == type(b)
    assert a.meta == b.meta

    assert len(a) == len(b)
    for row_a, row_b in zip(a, b):
        assert_array_equal(row_a, row_b)

    assert a.colnames == b.colnames
    for column_name in a.colnames:
        col_a = a[column_name]
        col_b = b[column_name]
        if isinstance(col_a, (Column, MaskedColumn)) and isinstance(col_b, (Column, MaskedColumn)):
            assert_description_equal(col_a.description, col_b.description)
            assert col_a.unit == col_b.unit
            assert col_a.meta == col_b.meta
            assert_array_equal(col_a.data, col_b.data)
            assert_array_equal(
                getattr(col_a, "mask", [False] * len(col_a)),
                getattr(col_b, "mask", [False] * len(col_b)),
            )


def assert_table_roundtrip(table, tmp_path):
    """
    Assert that a table can be written to an ASDF file and read back
    in without losing any of its essential properties.
    """
    file_path = tmp_path / "testable.asdf"

    with asdf.AsdfFile({"table": table}) as af:
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert_table_equal(table, af["table"])
        return af["table"]


def test_table(tmp_path):
    rows = [(1, 2.0, "x"), (4, 5.0, "y"), (5, 8.2, "z")]

    table = Table(rows=rows, names=("a", "b", "c"), dtype=("i4", "f8", "S1"))
    table.columns["a"].description = "RA"
    table.columns["a"].unit = "degree"
    table.columns["a"].meta = {"foo": "bar"}
    table.columns["c"].description = "Some description of some sort"

    assert_table_roundtrip(table, tmp_path)


def test_array_columns(tmp_path):
    data = np.array(
        [
            ([[1, 2], [3, 4]], 2.0, "x"),
            ([[5, 6], [7, 8]], 5.0, "y"),
            ([[9, 10], [11, 12]], 8.2, "z")
        ],
        dtype=[("a", "<i4", (2, 2)), ("b", "<f8"), ("c", "|S1")],
    )

    table = Table(data, copy=False)

    assert_table_roundtrip(table, tmp_path)


def test_structured_array_columns(tmp_path):
    rows = np.array(
        [
            ((1, "a"), 2.0, "x"),
            ((4, "b"), 5.0, "y"),
            ((5, "c"), 8.2, "z")
        ],
        dtype=[("a", [("a0", "<i4"), ("a1", "|S1")]), ("b", "<f8"), ("c", "|S1")]
    )

    table = Table(rows, copy=False)

    assert_table_roundtrip(table, tmp_path)


def test_table_row_order(tmp_path):
    data = np.array(
        [(1, 2.0, "x"), (4, 5.0, "y"),  (5, 8.2, "z")],
        dtype=[("a", "<i4"), ("b", "<f8"), ("c", "|S1")]
    )

    table = Table(data, copy=False)
    table.columns["a"].description = "RA"
    table.columns["a"].unit = "degree"
    table.columns["a"].meta = {"foo": "bar"}
    table.columns["c"].description = "Some description of some sort"

    assert_table_roundtrip(table, tmp_path)


def test_table_inline(tmp_path):
    rows = [(1, 2.0, "x"), (4, 5.0, "y"), (5, 8.2, "z")]
    table = Table(rows=rows, names=("a", "b", "c"), dtype=("i4", "f8", "S1"))
    table.columns["a"].description = "RA"
    table.columns["a"].unit = "degree"
    table.columns["a"].meta = {"foo": "bar"}
    table.columns["c"].description = "Some description of some sort"

    with asdf.config_context() as config:
        config.array_inline_threshold = 64
        assert_table_roundtrip(table, tmp_path)


def test_mismatched_columns():
    yaml = """
table: !<tag:astropy.org:astropy/table/table-1.0.0>
  columns:
  - !core/column-1.0.0
    data: !core/ndarray-1.0.0
      data: [0, 1, 2]
    name: a
  - !core/column-1.0.0
    data: !core/ndarray-1.0.0
      data: [0, 1, 2, 3]
    name: b
  colnames: [a, b]
    """

    buff = yaml_to_asdf(yaml)

    with pytest.raises(ValueError, match="Inconsistent data column lengths"):
        with asdf.open(buff):
            pass


def test_masked_table(tmp_path):
    rows = [(1, 2.0, "x"), (4, 5.0, "y"), (5, 8.2, "z")]
    table = Table(rows=rows, names=("a", "b", "c"), dtype=("i4", "f8", "S1"), masked=True)
    table.columns["a"].description = "RA"
    table.columns["a"].unit = "degree"
    table.columns["a"].meta = {"foo": "bar"}
    table.columns["a"].mask = [True, False, True]
    table.columns["c"].description = "Some description of some sort"

    assert_table_roundtrip(table, tmp_path)


def test_quantity_mixin(tmp_path):
    table = QTable()
    table["a"] = [1, 2, 3]
    table["b"] = ["x", "y", "z"]
    table["c"] = [2.0, 5.0, 8.2] * u.m

    assert_table_roundtrip(table, tmp_path)


def test_time_mixin(tmp_path):
    table = Table()
    table["a"] = [1, 2]
    table["b"] = ["x", "y"]
    table["c"] = Time(["2001-01-02T12:34:56", "2001-02-03T00:01:02"])

    result = assert_table_roundtrip(table, tmp_path)
    assert_time_equal(result["c"], table["c"])


def test_timedelta_mixin(tmp_path):
    table = Table()
    table["a"] = [1, 2]
    table["b"] = ["x", "y"]
    table["c"] = TimeDelta([1, 2] * u.day)

    result = assert_table_roundtrip(table, tmp_path)
    assert_time_delta_equal(result["c"], table["c"])


def test_skycoord_mixin(tmp_path):
    table = Table()
    table["a"] = [1, 2]
    table["b"] = ["x", "y"]
    table["c"] = SkyCoord([1, 2], [3, 4], unit="deg,deg", frame="fk4", obstime="J1990.5")

    # We can't use our assert_table_roundtrip helper because
    # astropy 4.0.x does not include an __eq__ method for SKyCoord.
    file_path = tmp_path / "test.asdf"
    with asdf.AsdfFile() as af:
        af["table"] = table
        af.write_to(file_path)

    with asdf.open(file_path) as af:
        assert_sky_coord_equal(af["table"]["c"], table["c"])


def test_earthlocation_mixin(tmp_path):
    table = Table()
    table["a"] = [1, 2]
    table["b"] = ["x", "y"]
    table["c"] = EarthLocation(x=[1, 2] * u.km, y=[3, 4] * u.km, z=[5, 6] * u.km)

    result = assert_table_roundtrip(table, tmp_path)
    assert_earth_location_equal(result["c"], table["c"])


def test_ndarray_mixin(tmp_path):
    table = Table()
    table["a"] = [1, 2]
    table["b"] = ["x", "y"]
    table["c"] = NdarrayMixin([5, 6])

    assert_table_roundtrip(table, tmp_path)


def test_asdf_table():
    yaml = """
table: !core/table-1.0.0
  columns:
    - !core/column-1.0.0
      data: !core/ndarray-1.0.0
        data: [1, 2, 3]
        datatype: float64
        shape: [3]
      description: RA
      meta: {foo: bar}
      name: a
      unit: !unit/unit-1.0.0 deg
    - !core/column-1.0.0
      data: !core/ndarray-1.0.0
        data: [4, 5, 6]
        datatype: float64
        shape: [3]
      description: DEC
      name: b
    - !core/column-1.0.0
      data: !core/ndarray-1.0.0
        data: [d, e, f]
        datatype: [ascii, 1]
        shape: [3]
      description: The target name
      name: c
    """
    buff = yaml_to_asdf(yaml)

    with asdf.open(buff) as af:
        table = af["table"]
        assert isinstance(table, Table)

        assert table["a"].name == "a"
        assert table["a"].description == "RA"
        assert table["a"].meta == {"foo": "bar"}
        assert table["a"].unit == u.deg
        assert_array_equal(table["a"].data, np.array([1, 2, 3], dtype=np.float64))

        assert table["b"].name == "b"
        assert table["b"].description == "DEC"
        assert_array_equal(table["b"].data, np.array([4, 5, 6], dtype=np.float64))

        assert table["c"].name == "c"
        assert table["c"].description == "The target name"
        assert_array_equal(table["c"].data, np.array([b"d", b"e", b"f"]))



from .core import TransformConverterBase


_MODEL_NAMES = [
    "AbsoluteUfunc",
    "AddUfunc",
    "ArccosUfunc",
    "ArccoshUfunc",
    "ArcsinUfunc",
    "ArcsinhUfunc",
    "Arctan2Ufunc",
    "ArctanUfunc",
    "ArctanhUfunc",
    "CbrtUfunc",
    "CosUfunc",
    "CoshUfunc",
    "Deg2radUfunc",
    # DivideUfunc is an alias for True_divideUfunc
    "DivmodUfunc",
    "Exp2Ufunc",
    "ExpUfunc",
    "Expm1Ufunc",
    "FabsUfunc",
    "Floor_divideUfunc",
    "FmodUfunc",
    "HypotUfunc",
    "Log10Ufunc",
    "Log1pUfunc",
    "Log2Ufunc",
    "LogUfunc",
    "Logaddexp2Ufunc",
    "LogaddexpUfunc",
    # ModUfunc is an alias for RemainderUfunc
    "MultiplyUfunc",
    "NegativeUfunc",
    "PositiveUfunc",
    "PowerUfunc",
    "Rad2degUfunc",
    "ReciprocalUfunc",
    "RemainderUfunc",
    "RintUfunc",
    "SinUfunc",
    "SinhUfunc",
    "SqrtUfunc",
    "SquareUfunc",
    "SubtractUfunc",
    "TanUfunc",
    "TanhUfunc",
    "True_divideUfunc",
]


class MathFunctionsConverter(TransformConverterBase):
    """
    ASDF support for serializing the math functions models,
    each of which corresponds to a numpy ufunc.
    """
    tags = ["tag:stsci.edu:asdf/transform/math_functions-*"]

    types = ["astropy.modeling.math_functions." + m for m in _MODEL_NAMES]

    def to_yaml_tree_transform(self, model, tag, ctx):
        return {"func_name": model.func.__name__}

    def from_yaml_tree_transform(self, node, tag, ctx):
        from astropy.modeling import math_functions

        klass_name = math_functions._make_class_name(node["func_name"])
        klass = getattr(math_functions, klass_name)
        return klass()



from asdf.extension import Converter
from asdf.tags.core.ndarray import NDArrayType


class QuantityConverter(Converter):
    tags = ["tag:stsci.edu:asdf/unit/quantity-*"]

    types = [
        "astropy.units.quantity.Quantity",
        # The Distance class has no tag of its own, so we
        # just serialize it as a quantity.
        "astropy.coordinates.distances.Distance",
    ]

    def to_yaml_tree(self, obj, tag, ctx):
        return {
            "value": obj.value,
            "unit": obj.unit,
        }

    def from_yaml_tree(self, node, tag, ctx):
        from astropy.units import Quantity

        value = node["value"]
        if isinstance(value, NDArrayType):
            # TODO: Why doesn't NDArrayType work?  This needs some research
            # and documentation.
            value = value._make_array()

        return Quantity(value, unit=node["unit"])



from asdf.extension import Converter


class UnitConverter(Converter):
    tags = ["tag:stsci.edu:asdf/unit/unit-*"]

    types = [
        "astropy.units.core.CompositeUnit",
        "astropy.units.core.IrreducibleUnit",
        "astropy.units.core.NamedUnit",
        "astropy.units.core.PrefixUnit",
        "astropy.units.core.Unit",
        "astropy.units.core.UnitBase",
        "astropy.units.core.UnrecognizedUnit",
        "astropy.units.function.mixin.IrreducibleFunctionUnit",
        "astropy.units.function.mixin.RegularFunctionUnit"
    ]

    def to_yaml_tree(self, obj, tag, ctx):
        from astropy.units import UnitsError

        try:
            return obj.to_string(format="vounit")
        except (UnitsError, ValueError) as e:
            raise ValueError(f"Unit '{obj}' is not representable as VOUnit and cannot be serialized to ASDF") from e

    def from_yaml_tree(self, node, tag, ctx):
        from astropy.units import Unit

        return Unit(node, format="vounit", parse_strict="silent")
